
-module(wormhole_settle_tx).

-include("channel_txs.hrl").

-behavior(aetx).

%% Behavior API
-export([new/1,
         type/0,
         fee/1,
         ttl/1,
         nonce/1,
         origin/1,
         check/5,
         process/5,
         accounts/1,
         signers/2,
         serialization_template/1,
         serialize/1,
         deserialize/2,
         for_client/1
        ]).

%%%===================================================================
%%% Types
%%%===================================================================

-define(CHANNEL_SETTLE_TX_VSN, 1).
-define(CHANNEL_SETTLE_TX_TYPE, channel_settle_tx).
-define(CHANNEL_SETTLE_TX_FEE, 4).

-opaque tx() :: #channel_settle_tx{}.

-export_type([tx/0]).

%%%===================================================================
%%% Behaviour API
%%%===================================================================

-spec new(map()) -> {ok, aetx:tx()}.
new(#{channel_id        := ChannelId,
      from              := FromPubKey,
      initiator_amount  := InitiatorAmount,
      responder_amount  := ResponderAmount,
      ttl               := TTL,
      fee               := Fee,
      nonce             := Nonce}) ->
    Tx = #channel_settle_tx{
            channel_id        = ChannelId,
            from              = FromPubKey,
            initiator_amount  = InitiatorAmount,
            responder_amount  = ResponderAmount,
            ttl               = TTL,
            fee               = Fee,
            nonce             = Nonce},
    {ok, aetx:new(?MODULE, Tx)}.

type() ->
    ?CHANNEL_SETTLE_TX_TYPE.

-spec fee(tx()) -> non_neg_integer().
fee(#channel_settle_tx{fee = Fee}) ->
    Fee.

-spec ttl(tx()) -> aec_blocks:height().
ttl(#channel_settle_tx{ttl = TTL}) ->
    TTL.

-spec nonce(tx()) -> non_neg_integer().
nonce(#channel_settle_tx{nonce = Nonce}) ->
    Nonce.

-spec origin(tx()) -> aec_keys:pubkey().
origin(#channel_settle_tx{from = FromPubKey}) ->
    FromPubKey.

-spec check(tx(), aetx:tx_context(), aec_trees:trees(), aec_blocks:height(), non_neg_integer()) ->
        {ok, aec_trees:trees()} | {error, term()}.
check(#channel_settle_tx{channel_id       = ChannelId,
                         from             = FromPubKey,
                         initiator_amount = InitiatorAmount,
                         responder_amount = ResponderAmount,
                         fee              = Fee,
                         nonce            = Nonce}, _Context, Trees, Height, _ConsensusVersion) ->
    Checks =
        [fun() -> aetx_utils:check_account(FromPubKey, Trees, Nonce, Fee) end,
         fun() -> check_channel(ChannelId, FromPubKey, InitiatorAmount,
                                ResponderAmount, Height, Trees) end],
    case aeu_validation:run(Checks) of
        ok ->
            {ok, Trees};
        {error, _Reason} = Error ->
            Error
    end.

-spec process(tx(), aetx:tx_context(), aec_trees:trees(), aec_blocks:height(), non_neg_integer()) ->
        {ok, aec_trees:trees()}.
process(#channel_settle_tx{channel_id       = ChannelId,
                           from             = FromPubKey,
                           initiator_amount = InitiatorAmount,
                           responder_amount = ResponderAmount,
                           fee              = Fee,
                           nonce            = Nonce}, _Context, Trees, _Height, _ConsensusVersion) ->
    AccountsTree0 = aec_trees:accounts(Trees),
    ChannelsTree0 = aec_trees:channels(Trees),

    Channel0        = wormhole_state_tree:get(ChannelId, ChannelsTree0),
    Initiator       = wormhole_channels:initiator(Channel0),
    InitiatorAmount = wormhole_channels:initiator_amount(Channel0), % same amt
    TotalAmount     = wormhole_channels:total_amount(Channel0),
    Responder       = wormhole_channels:responder(Channel0),
    ResponderAmount = TotalAmount - InitiatorAmount, % same amt

    InitiatorAccount0       = aec_accounts_trees:get(Initiator, AccountsTree0),
    ResponderAccount0       = aec_accounts_trees:get(Responder, AccountsTree0),
    {ok, InitiatorAccount1} = aec_accounts:earn(InitiatorAccount0, InitiatorAmount),
    {ok, ResponderAccount1} = aec_accounts:earn(ResponderAccount0, ResponderAmount),

    {InitiatorAccount3, ResponderAccount3} =
        case FromPubKey of
            Initiator ->
                {ok, InitiatorAccount2} = aec_accounts:spend(InitiatorAccount1, Fee, Nonce),
                {InitiatorAccount2, ResponderAccount1};
            Responder ->
                {ok, ResponderAccount2} = aec_accounts:spend(ResponderAccount1, Fee, Nonce),
                {InitiatorAccount1, ResponderAccount2}
        end,

    AccountsTree1 = aec_accounts_trees:enter(InitiatorAccount3, AccountsTree0),
    AccountsTree2 = aec_accounts_trees:enter(ResponderAccount3, AccountsTree1),

    ChannelsTree1 = wormhole_state_tree:delete(wormhole_channels:id(Channel0), ChannelsTree0),

    Trees1 = aec_trees:set_accounts(Trees, AccountsTree2),
    Trees2 = aec_trees:set_channels(Trees1, ChannelsTree1),
    {ok, Trees2}.

-spec accounts(tx()) -> list(aec_keys:pubkey()).
accounts(#channel_settle_tx{channel_id = ChannelId}) ->
    case aec_chain:get_channel(ChannelId) of
        {ok, Channel} ->
            [wormhole_channels:initiator(Channel), wormhole_channels:responder(Channel)];
        {error, not_found} -> []
    end.

-spec signers(tx(), aec_trees:trees()) -> {ok, list(aec_keys:pubkey())}.
signers(#channel_settle_tx{from = FromPubKey}, _) ->
    {ok, [FromPubKey]}.

-spec serialize(tx()) -> {vsn(), list()}.
serialize(#channel_settle_tx{channel_id       = ChannelId,
                             from             = FromPubKey,
                             initiator_amount = InitiatorAmount,
                             responder_amount = ResponderAmount,
                             ttl              = TTL,
                             fee              = Fee,
                             nonce            = Nonce}) ->
    {version(),
    [ {channel_id       , ChannelId}
    , {from             , FromPubKey}
    , {initiator_amount , InitiatorAmount}
    , {responder_amount , ResponderAmount}
    , {ttl              , TTL}
    , {fee              , Fee}
    , {nonce            , Nonce}
    ]}.

-spec deserialize(vsn(), list()) -> tx().
deserialize(?CHANNEL_SETTLE_TX_VSN,
            [ {channel_id       , ChannelId}
            , {from             , FromPubKey}
            , {initiator_amount , InitiatorAmount}
            , {responder_amount , ResponderAmount}
            , {ttl              , TTL}
            , {fee              , Fee}
            , {nonce            , Nonce}]) ->
    #channel_settle_tx{channel_id       = ChannelId,
                       from             = FromPubKey,
                       initiator_amount = InitiatorAmount,
                       responder_amount = ResponderAmount,
                       ttl              = TTL,
                       fee              = Fee,
                       nonce            = Nonce}.

-spec for_client(tx()) -> map().
for_client(#channel_settle_tx{channel_id       = ChannelId,
                              from             = FromPubKey,
                              initiator_amount = InitiatorAmount,
                              responder_amount = ResponderAmount,
                              ttl              = TTL,
                              fee              = Fee,
                              nonce            = Nonce}) ->
    #{<<"data_schema">>      => <<"ChannelSettleTxJSON">>, % swagger schema name
      <<"vsn">>              => version(),
      <<"channel_id">>       => aec_base58c:encode(channel, ChannelId),
      <<"from">>             => aec_base58c:encode(account_pubkey, FromPubKey),
      <<"initiator_amount">> => InitiatorAmount,
      <<"responder_amount">> => ResponderAmount,
      <<"ttl">>              => TTL,
      <<"fee">>              => Fee,
      <<"nonce">>            => Nonce}.


serialization_template(?CHANNEL_SETTLE_TX_VSN) ->
    [ {channel_id       , binary}
    , {from             , binary}
    , {initiator_amount , int}
    , {responder_amount , int}
    , {ttl              , int}
    , {fee              , int}
    , {nonce            , int}
    ].

%%%===================================================================
%%% Internal functions
%%%===================================================================

check_channel(ChannelId, FromPubKey, InitiatorAmount, ResponderAmount, Height, Trees) ->
    ChannelsTree = aec_trees:channels(Trees),
    case wormhole_state_tree:lookup(ChannelId, ChannelsTree) of
        none ->
            {error, channel_does_not_exist};
        {value, Channel} ->
            Checks =
                [fun() -> wormhole_utils:check_is_peer(FromPubKey, wormhole_channels:peers(Channel)) end,
                 fun() -> check_solo_closed(Channel, Height) end,
                 %% check total amount
                 fun() -> wormhole_utils:check_are_funds_in_channel(ChannelId,
                                InitiatorAmount + ResponderAmount, Trees) end,
                 %% check individual amounts are what is expected
                 fun() -> check_peer_amount(InitiatorAmount,
                                            wormhole_channels:initiator_amount(Channel))
                 end,
                 fun() -> check_peer_amount(ResponderAmount,
                                            wormhole_channels:responder_amount(Channel))
                 end],
            aeu_validation:run(Checks)
    end.

check_peer_amount(ExpectedAmt, Amt) ->
    case Amt =:= ExpectedAmt of
        true -> ok;
        false -> {error, wrong_amt}
    end.

check_solo_closed(Channel, Height) ->
    case wormhole_channels:is_solo_closed(Channel, Height) of
        true  -> ok;
        false -> {error, channel_not_closed}
    end.

-spec version() -> non_neg_integer().
version() ->
    ?CHANNEL_SETTLE_TX_VSN.
